@@ -1,7 +1,7 @@
 #!/usr/bin/env python
 u"""
 utilities.py
-Written by Tyler Sutterley (04/2022)
+Written by Tyler Sutterley (12/2022)
 Download and management utilities for syncing time and auxiliary files
 
 PYTHON DEPENDENCIES:
@@ -9,6 +9,7 @@ PYTHON DEPENDENCIES:
         https://pypi.python.org/pypi/lxml
 
 UPDATE HISTORY:
+    Updated 12/2022: functions for managing and maintaining git repositories
     Updated 04/2022: updated docstrings to numpy documentation format
     Updated 10/2021: using python logging for handling verbose output
     Updated 03/2021: added sha1 option for retrieving file hashes
@@ -29,8 +30,10 @@ import socket
 import inspect
 import hashlib
 import logging
+import warnings
 import posixpath
 import lxml.etree
+import subprocess
 import calendar,time
 if sys.version_info[0] == 2:
     from urllib import quote_plus
@@ -41,7 +44,7 @@ else:
     from http.cookiejar import CookieJar
     import urllib.request as urllib2
 
-#-- PURPOSE: get absolute path within a package from a relative path
+# PURPOSE: get absolute path within a package from a relative path
 def get_data_path(relpath):
     """
     Get the absolute path within a package from a relative path
@@ -51,16 +54,16 @@ def get_data_path(relpath):
     relpath: str,
         relative path
     """
-    #-- current file path
+    # current file path
     filename = inspect.getframeinfo(inspect.currentframe()).filename
     filepath = os.path.dirname(os.path.abspath(filename))
     if isinstance(relpath,list):
-        #-- use *splat operator to extract from list
+        # use *splat operator to extract from list
         return os.path.join(filepath,*relpath)
     elif isinstance(relpath,str):
         return os.path.join(filepath,relpath)
 
-#-- PURPOSE: get the hash value of a file
+# PURPOSE: get the hash value of a file
 def get_hash(local, algorithm='MD5'):
     """
     Get the hash value from a local file or BytesIO object
@@ -75,17 +78,17 @@ def get_hash(local, algorithm='MD5'):
             - ``'MD5'``: Message Digest
             - ``'sha1'``: Secure Hash Algorithm
     """
-    #-- check if open file object or if local file exists
+    # check if open file object or if local file exists
     if isinstance(local, io.IOBase):
         if (algorithm == 'MD5'):
             return hashlib.md5(local.getvalue()).hexdigest()
         elif (algorithm == 'sha1'):
             return hashlib.sha1(local.getvalue()).hexdigest()
     elif os.access(os.path.expanduser(local),os.F_OK):
-        #-- generate checksum hash for local file
-        #-- open the local_file in binary read mode
+        # generate checksum hash for local file
+        # open the local_file in binary read mode
         with open(os.path.expanduser(local), 'rb') as local_buffer:
-            #-- generate checksum hash for a given type
+            # generate checksum hash for a given type
             if (algorithm == 'MD5'):
                 return hashlib.md5(local_buffer.read()).hexdigest()
             elif (algorithm == 'sha1'):
@@ -93,7 +96,44 @@ def get_hash(local, algorithm='MD5'):
     else:
         return ''
 
-#-- PURPOSE: recursively split a url path
+# PURPOSE: get the git hash value
+def get_git_revision_hash(refname='HEAD', short=False):
+    """
+    Get the git hash value for a particular reference
+
+    Parameters
+    ----------
+    refname: str, default HEAD
+        Symbolic reference name
+    short: bool, default False
+        Return the shorted hash value
+    """
+    # get path to .git directory from current file path
+    filename = inspect.getframeinfo(inspect.currentframe()).filename
+    basepath = os.path.dirname(os.path.dirname(os.path.abspath(filename)))
+    gitpath = os.path.join(basepath,'.git')
+    # build command
+    cmd = ['git', f'--git-dir={gitpath}', 'rev-parse']
+    cmd.append('--short') if short else None
+    cmd.append(refname)
+    # get output
+    with warnings.catch_warnings():
+        return str(subprocess.check_output(cmd), encoding='utf8').strip()
+
+# PURPOSE: get the current git status
+def get_git_status():
+    """Get the status of a git repository as a boolean value
+    """
+    # get path to .git directory from current file path
+    filename = inspect.getframeinfo(inspect.currentframe()).filename
+    basepath = os.path.dirname(os.path.dirname(os.path.abspath(filename)))
+    gitpath = os.path.join(basepath,'.git')
+    # build command
+    cmd = ['git', f'--git-dir={gitpath}', 'status', '--porcelain']
+    with warnings.catch_warnings():
+        return bool(subprocess.check_output(cmd))
+
+# PURPOSE: recursively split a url path
 def url_split(s):
     """
     Recursively split a url path into a list
@@ -110,7 +150,7 @@ def url_split(s):
         return tail,
     return url_split(head) + (tail,)
 
-#-- PURPOSE: convert file lines to arguments
+# PURPOSE: convert file lines to arguments
 def convert_arg_line_to_args(arg_line):
     """
     Convert file lines to arguments
@@ -120,13 +160,13 @@ def convert_arg_line_to_args(arg_line):
     arg_line: str
         line string containing a single argument and/or comments
     """
-    #-- remove commented lines and after argument comments
+    # remove commented lines and after argument comments
     for arg in re.sub(r'\#(.*?)$',r'',arg_line).split():
         if not arg.strip():
             continue
         yield arg
 
-#-- PURPOSE: returns the Unix timestamp value for a formatted date string
+# PURPOSE: returns the Unix timestamp value for a formatted date string
 def get_unix_time(time_string, format='%Y-%m-%d %H:%M:%S'):
     """
     Get the Unix timestamp value for a formatted date string
@@ -145,7 +185,7 @@ def get_unix_time(time_string, format='%Y-%m-%d %H:%M:%S'):
     else:
         return calendar.timegm(parsed_time)
 
-#-- PURPOSE: rounds a number to an even number less than or equal to original
+# PURPOSE: rounds a number to an even number less than or equal to original
 def even(value):
     """
     Rounds a number to an even number less than or equal to original
@@ -157,7 +197,7 @@ def even(value):
     """
     return 2*int(value//2)
 
-#-- PURPOSE: rounds a number upward to its nearest integer
+# PURPOSE: rounds a number upward to its nearest integer
 def ceil(value):
     """
     Rounds a number upward to its nearest integer
@@ -169,7 +209,7 @@ def ceil(value):
     """
     return -int(-value//1)
 
-#-- PURPOSE: make a copy of a file with all system information
+# PURPOSE: make a copy of a file with all system information
 def copy(source, destination, move=False, **kwargs):
     """
     Copy or move a file with all system information
@@ -185,14 +225,14 @@ def copy(source, destination, move=False, **kwargs):
     """
     source = os.path.abspath(os.path.expanduser(source))
     destination = os.path.abspath(os.path.expanduser(destination))
-    #-- log source and destination
-    logging.info('{0} -->\n\t{1}'.format(source,destination))
+    # log source and destination
+    logging.info(f'{source} -->\n\t{destination}')
     shutil.copyfile(source, destination)
     shutil.copystat(source, destination)
     if move:
         os.remove(source)
 
-#-- PURPOSE: check ftp connection
+# PURPOSE: check ftp connection
 def check_ftp_connection(HOST, username=None, password=None):
     """
     Check internet connection with ftp host
@@ -206,7 +246,7 @@ def check_ftp_connection(HOST, username=None, password=None):
     password: str or NoneType
         ftp password
     """
-    #-- attempt to connect to ftp host
+    # attempt to connect to ftp host
     try:
         f = ftplib.FTP(HOST)
         f.login(username, password)
@@ -218,7 +258,7 @@ def check_ftp_connection(HOST, username=None, password=None):
     else:
         return True
 
-#-- PURPOSE: list a directory on a ftp host
+# PURPOSE: list a directory on a ftp host
 def ftp_list(HOST, username=None, password=None, timeout=None,
     basename=False, pattern=None, sort=False):
     """
@@ -248,52 +288,52 @@ def ftp_list(HOST, username=None, password=None, timeout=None,
     mtimes: list
         last modification times for items in the directory
     """
-    #-- verify inputs for remote ftp host
+    # verify inputs for remote ftp host
     if isinstance(HOST, str):
         HOST = url_split(HOST)
-    #-- try to connect to ftp host
+    # try to connect to ftp host
     try:
         ftp = ftplib.FTP(HOST[0],timeout=timeout)
     except (socket.gaierror,IOError):
-        raise RuntimeError('Unable to connect to {0}'.format(HOST[0]))
+        raise RuntimeError(f'Unable to connect to {HOST[0]}')
     else:
         ftp.login(username,password)
-        #-- list remote path
+        # list remote path
         output = ftp.nlst(posixpath.join(*HOST[1:]))
-        #-- get last modified date of ftp files and convert into unix time
+        # get last modified date of ftp files and convert into unix time
         mtimes = [None]*len(output)
-        #-- iterate over each file in the list and get the modification time
+        # iterate over each file in the list and get the modification time
         for i,f in enumerate(output):
             try:
-                #-- try sending modification time command
-                mdtm = ftp.sendcmd('MDTM {0}'.format(f))
+                # try sending modification time command
+                mdtm = ftp.sendcmd(f'MDTM {f}')
             except ftplib.error_perm:
-                #-- directories will return with an error
+                # directories will return with an error
                 pass
             else:
-                #-- convert the modification time into unix time
+                # convert the modification time into unix time
                 mtimes[i] = get_unix_time(mdtm[4:], format="%Y%m%d%H%M%S")
-        #-- reduce to basenames
+        # reduce to basenames
         if basename:
             output = [posixpath.basename(i) for i in output]
-        #-- reduce using regular expression pattern
+        # reduce using regular expression pattern
         if pattern:
             i = [i for i,f in enumerate(output) if re.search(pattern,f)]
-            #-- reduce list of listed items and last modified times
+            # reduce list of listed items and last modified times
             output = [output[indice] for indice in i]
             mtimes = [mtimes[indice] for indice in i]
-        #-- sort the list
+        # sort the list
         if sort:
             i = [i for i,j in sorted(enumerate(output), key=lambda i: i[1])]
-            #-- sort list of listed items and last modified times
+            # sort list of listed items and last modified times
             output = [output[indice] for indice in i]
             mtimes = [mtimes[indice] for indice in i]
-        #-- close the ftp connection
+        # close the ftp connection
         ftp.close()
-        #-- return the list of items and last modified times
+        # return the list of items and last modified times
         return (output,mtimes)
 
-#-- PURPOSE: download a file from a ftp host
+# PURPOSE: download a file from a ftp host
 def from_ftp(HOST, username=None, password=None, timeout=None,
     local=None, hash='', chunk=8192, verbose=False, fid=sys.stdout,
     mode=0o775):
@@ -328,59 +368,59 @@ def from_ftp(HOST, username=None, password=None, timeout=None,
     remote_buffer: obj
         BytesIO representation of file
     """
-    #-- create logger
+    # create logger
     loglevel = logging.INFO if verbose else logging.CRITICAL
     logging.basicConfig(stream=fid, level=loglevel)
-    #-- verify inputs for remote ftp host
+    # verify inputs for remote ftp host
     if isinstance(HOST, str):
         HOST = url_split(HOST)
-    #-- try downloading from ftp
+    # try downloading from ftp
     try:
-        #-- try to connect to ftp host
+        # try to connect to ftp host
         ftp = ftplib.FTP(HOST[0],timeout=timeout)
     except (socket.gaierror,IOError):
-        raise RuntimeError('Unable to connect to {0}'.format(HOST[0]))
+        raise RuntimeError(f'Unable to connect to {HOST[0]}')
     else:
         ftp.login(username,password)
-        #-- remote path
+        # remote path
         ftp_remote_path = posixpath.join(*HOST[1:])
-        #-- copy remote file contents to bytesIO object
+        # copy remote file contents to bytesIO object
         remote_buffer = io.BytesIO()
-        ftp.retrbinary('RETR {0}'.format(ftp_remote_path),
+        ftp.retrbinary(f'RETR {ftp_remote_path}',
             remote_buffer.write, blocksize=chunk)
         remote_buffer.seek(0)
-        #-- save file basename with bytesIO object
+        # save file basename with bytesIO object
         remote_buffer.filename = HOST[-1]
-        #-- generate checksum hash for remote file
+        # generate checksum hash for remote file
         remote_hash = hashlib.md5(remote_buffer.getvalue()).hexdigest()
-        #-- get last modified date of remote file and convert into unix time
-        mdtm = ftp.sendcmd('MDTM {0}'.format(ftp_remote_path))
+        # get last modified date of remote file and convert into unix time
+        mdtm = ftp.sendcmd(f'MDTM {ftp_remote_path}')
         remote_mtime = get_unix_time(mdtm[4:], format="%Y%m%d%H%M%S")
-        #-- compare checksums
+        # compare checksums
         if local and (hash != remote_hash):
-            #-- convert to absolute path
+            # convert to absolute path
             local = os.path.abspath(local)
-            #-- create directory if non-existent
+            # create directory if non-existent
             if not os.access(os.path.dirname(local), os.F_OK):
                 os.makedirs(os.path.dirname(local), mode)
-            #-- print file information
+            # print file information
             args = (posixpath.join(*HOST),local)
             logging.info('{0} -->\n\t{1}'.format(*args))
-            #-- store bytes to file using chunked transfer encoding
+            # store bytes to file using chunked transfer encoding
             remote_buffer.seek(0)
             with open(os.path.expanduser(local), 'wb') as f:
                 shutil.copyfileobj(remote_buffer, f, chunk)
-            #-- change the permissions mode
+            # change the permissions mode
             os.chmod(local,mode)
-            #-- keep remote modification time of file and local access time
+            # keep remote modification time of file and local access time
             os.utime(local, (os.stat(local).st_atime, remote_mtime))
-        #-- close the ftp connection
+        # close the ftp connection
         ftp.close()
-        #-- return the bytesIO object
+        # return the bytesIO object
         remote_buffer.seek(0)
         return remote_buffer
 
-#-- PURPOSE: check internet connection
+# PURPOSE: check internet connection
 def check_connection(HOST):
     """
     Check internet connection with http host
@@ -390,7 +430,7 @@ def check_connection(HOST):
     HOST: str
         remote http host
     """
-    #-- attempt to connect to http host
+    # attempt to connect to http host
     try:
         urllib2.urlopen(HOST,timeout=20,context=ssl.SSLContext())
     except urllib2.URLError:
@@ -398,7 +438,7 @@ def check_connection(HOST):
     else:
         return True
 
-#-- PURPOSE: download a file from a http host
+# PURPOSE: download a file from a http host
 def from_http(HOST, timeout=None, context=ssl.SSLContext(),
     local=None, hash='', chunk=16384, verbose=False, fid=sys.stdout,
     mode=0o775):
@@ -433,50 +473,50 @@ def from_http(HOST, timeout=None, context=ssl.SSLContext(),
     remote_buffer: obj
         BytesIO representation of file
     """
-    #-- create logger
+    # create logger
     loglevel = logging.INFO if verbose else logging.CRITICAL
     logging.basicConfig(stream=fid, level=loglevel)
-    #-- verify inputs for remote http host
+    # verify inputs for remote http host
     if isinstance(HOST, str):
         HOST = url_split(HOST)
-    #-- try downloading from http
+    # try downloading from http
     try:
-        #-- Create and submit request.
+        # Create and submit request.
         request = urllib2.Request(posixpath.join(*HOST))
         response = urllib2.urlopen(request,timeout=timeout,context=context)
     except (urllib2.HTTPError, urllib2.URLError):
         raise Exception('Download error from {0}'.format(posixpath.join(*HOST)))
     else:
-        #-- copy remote file contents to bytesIO object
+        # copy remote file contents to bytesIO object
         remote_buffer = io.BytesIO()
         shutil.copyfileobj(response, remote_buffer, chunk)
         remote_buffer.seek(0)
-        #-- save file basename with bytesIO object
+        # save file basename with bytesIO object
         remote_buffer.filename = HOST[-1]
-        #-- generate checksum hash for remote file
+        # generate checksum hash for remote file
         remote_hash = hashlib.md5(remote_buffer.getvalue()).hexdigest()
-        #-- compare checksums
+        # compare checksums
         if local and (hash != remote_hash):
-            #-- convert to absolute path
+            # convert to absolute path
             local = os.path.abspath(local)
-            #-- create directory if non-existent
+            # create directory if non-existent
             if not os.access(os.path.dirname(local), os.F_OK):
                 os.makedirs(os.path.dirname(local), mode)
-            #-- print file information
+            # print file information
             args = (posixpath.join(*HOST),local)
             logging.info('{0} -->\n\t{1}'.format(*args))
-            #-- store bytes to file using chunked transfer encoding
+            # store bytes to file using chunked transfer encoding
             remote_buffer.seek(0)
             with open(os.path.expanduser(local), 'wb') as f:
                 shutil.copyfileobj(remote_buffer, f, chunk)
-            #-- change the permissions mode
+            # change the permissions mode
             os.chmod(local,mode)
-        #-- return the bytesIO object
+        # return the bytesIO object
         remote_buffer.seek(0)
         return remote_buffer
 
-#-- PURPOSE: list a directory on the GFZ ICGEM https server
-#-- http://icgem.gfz-potsdam.de
+# PURPOSE: list a directory on the GFZ ICGEM https server
+# http://icgem.gfz-potsdam.de
 def icgem_list(host='http://icgem.gfz-potsdam.de/tom_longtime',timeout=None,
     parser=lxml.etree.HTMLParser()):
     """
@@ -498,17 +538,17 @@ def icgem_list(host='http://icgem.gfz-potsdam.de/tom_longtime',timeout=None,
     colfiles: dict
         Static gravity field file urls mapped by field name
     """
-    #-- try listing from https
+    # try listing from https
     try:
-        #-- Create and submit request.
+        # Create and submit request.
         request = urllib2.Request(host)
         tree = lxml.etree.parse(urllib2.urlopen(request,timeout=timeout),parser)
     except:
-        raise Exception('List error from {0}'.format(host))
+        raise Exception(f'List error from {host}')
     else:
-        #-- read and parse request for files
+        # read and parse request for files
         colfiles = tree.xpath('//td[@class="tom-cell-modelfile"]//a/@href')
-        #-- reduce list of files to find gfc files
-        #-- return the dict of model files mapped by name
+        # reduce list of files to find gfc files
+        # return the dict of model files mapped by name
         return {re.findall('(.*?).gfc',posixpath.basename(f)).pop():url_split(f)
             for i,f in enumerate(colfiles) if re.search('gfc$',f)}
