@@ -8,4 +8,6 @@ from pkg_resources import get_distribution
 # get version
 version = get_distribution("geoid_toolkit").version
 # append "v" before the version
-full_version = "v{0}".format(version)
+full_version = f"v{version}"
+# get project name
+project_name = get_distribution("geoid_toolkit").project_name
